from abc import ABC, abstractmethod
from sys import platform
import socket, threading
import logging
import traceback
import time

log = logging.getLogger(__name__)


class TCPServerBase(ABC):
    """
    Base class to host common TCP server methods.
    """

    def __init__(self):
        # create a multithreaded event, which is basically a
        # thread-safe boolean
        self._is_running = threading.Event()

        # this socket will be used to listen to incoming connections
        self._socket = None

        # this will contain the shell for the connected client.
        # we don't yet initialize it, since we need to get the
        # stdin and stdout objects after the connection is made.
        self.client_shell = None

        # this will contain the thread that will listen for incoming
        # connections and data.
        self._listen_thread = None

        # list of active connections maintained by this server
        self._connection_threads = []

    # To start the server, we open the socket and create
    # the listening thread.
    def start(self, address=None, port=None, timeout=None):
        """Start Server"""
        address = address or self.address
        port = port or self.port
        timeout = timeout or self.timeout

        if not self._is_running.is_set():
            self._is_running.set()

            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

            # reuse port is not avaible on windows
            if platform == "linux" or platform == "linux2":
                self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, True)

            self._socket.settimeout(timeout)
            self._socket.bind((address, port))

            self._listen_thread = threading.Thread(target=self._listen)
            self._listen_thread.start()

    # To stop the server, we must join the listen thread
    # and close the socket.
    def stop(self):
        """Stop Server"""
        if self._is_running.is_set():
            self._is_running.clear()
            self._listen_thread.join()
            self._socket.close()
            for connection_thread in self._connection_threads:
                connection_thread.join()

    # The listen function will constantly run if the server is running.
    # We wait for a connection, if a connection is made, we will call
    # our connection function.
    def _listen(self):
        while self._is_running.is_set():
            try:
                self._socket.listen()
                client, addr = self._socket.accept()
                connection_thread = threading.Thread(
                    target=self.connection_function, args=(client,)
                )
                connection_thread.start()
                self._connection_threads.append(connection_thread)
            except socket.timeout:
                pass
            time.sleep(0.01)

    @abstractmethod
    def connection_function(self, client):
        pass



import yaml
import importlib.util
import logging

from fakenos.core.pydantic_models import model_nos_attributes

log = logging.getLogger(__name__)


class Nos:
    """
    Base class to build NOS plugins instances to use with FakeNOS.
    """

    def __init__(
        self, name: str = None, commands: dict = None, initial_prompt: str = None
    ) -> None:
        """
        Method to instantiate Nos Instance

        :param name: NOS plugin name
        :param commands: dictionary of NOS commands
        :param initial_prompt: NOS initial prompt
        """
        self.name = name or "noname"
        self.commands = commands or {}
        self.initial_prompt = initial_prompt or "noprompt"
        
        # make sure to validate NOS attributes if Nos class instantiated with them
        if commands or name or initial_prompt:
            self.validate()
        
    def validate(self) -> None:
        """
        Method to validate NOS attributes: commands, name, initial prompt - using 
        Pydantic models, raises ValidationError on failure.
        """
        _ = model_nos_attributes(
            name=self.name,
            initial_prompt=self.initial_prompt,
            commands=self.commands,
        )
        log.debug(f"{self.name} NOS attributes validation succeeded")
        
    def from_dict(self, data: dict) -> None:
        """
        Method to build NOS from dictionary data.

        Sample NOS dictionary::

            nos_plugin_dict = {
                "name": "MyFakeNOSPlugin",
                "initial_prompt": "{base_prompt}>",
                "commands": {
                    "terminal width 511": {"output": "", "help": "Set terminal width to 511"},
                    "terminal length 0": {"output": "", "help": "Set terminal length to 0"},
                    "show clock": {"output": "MyFakeNOSPlugin system time is 00:00:00"},
                },
            }

        :param data: NOS dictionary
        """
        self.name = data.get("name", self.name)
        self.commands = data.get("commands", self.commands)
        self.initial_prompt = data.get("initial_prompt", self.initial_prompt)
        self.validate()
        
    def from_yaml(self, data: str) -> None:
        """
        Method to build NOS from YAML data.

        Sample NOS YAML file content::

            name: "MyFakeNOSPlugin"
            initial_prompt: "{base_prompt}>"
            commands:
              terminal width 511: {"output": "", "help": "Set terminal width to 511"}
              terminal length 0: {"output": "", "help": "Set terminal length to 0"}
              show clock: {"output": "MyFakeNOSPlugin system time is 00:00:00"}

        :param data: YAML structured text
        """
        self.from_dict(yaml.safe_load(data))
        self.validate()
        
    def from_module(self, data: str) -> None:
        """
        Method to import NOS data from python file.

        Sample Python NOS plugin file::

            name = "MyFakeNOSPlugin"

            initial_prompt = "{base_prompt}>"

            commands = {
                "terminal width 511": {"output": "", "help": "Set terminal width to 511"},
                "terminal length 0": {"output": "", "help": "Set terminal length to 0"},
                "show clock": {"output": "MyFakeNOSPlugin system time is 00:00:00"},
            }

        :param data: OS path string to Python .py file
        """
        # load .py file using this recipe
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        spec = importlib.util.spec_from_file_location("nos_module", data)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        # source attributes from loaded .py module
        self.name = getattr(module, "name", self.name)
        self.commands = getattr(module, "commands", self.commands)
        self.initial_prompt = getattr(module, "initial_prompt", self.initial_prompt)
        self.validate()
        
    def from_file(self, data: str) -> None:
        """
        Method to load NOS from YAML or Python file

        :param data: OS path string to `.yaml/.yml` or `.py` file with NOS data
        """
        if data.endswith(".yaml") or data.endswith(".yml"):
            with open(data, "r", encoding="utf-8") as f:
                self.from_yaml(f.read())
        elif data.endswith(".py"):
            self.from_module(data)
        else:
            raise ValueError(
                f"Unsupported '{data}' file extension, supported .py, .yml, .yaml"
            )



"""
FakeNOS Command Line Tool for running fake servers.
"""

import argparse
import time

from fakenos import FakeNOS

__version__ = "0.1.0"

# form argparser menu:
description_text = """-i --inventory   OS Path to inventory file
"""

argparser = argparse.ArgumentParser(
    description="FakeNOS, version {}".format(__version__),
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
opts = argparser.add_argument_group(description=description_text)

opts.add_argument(
    "-i",
    "--inventory",
    action="store",
    dest="INVENTORY",
    default=None,
    type=str,
    help=argparse.SUPPRESS,
)


# extract argparser arguments:
args = argparser.parse_args()
INVENTORY = args.INVENTORY


def run_cli():
    """Function to start FakeNOS CLI"""
    net = FakeNOS(inventory=INVENTORY)
    net.start()
    print("Started servers:")
    print(net.list_hosts())


if __name__ == "__main__":
    run_cli()



"""
File to contain pydantic models for plugins input/output data validation
"""
from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictFloat,
    StrictStr,
    conlist,
    IPvAnyAddress,
    conint,
    root_validator,
)
from typing import Union, Optional, List, Any, Dict, Callable, Tuple

try:
    from typing import Literal  # works with >=py3.8
except ImportError:
    from typing_extensions import Literal  # works with <py3.8


# ---------------------------------------------------------------------------------------
# FakeNOS inventory data model components
# ---------------------------------------------------------------------------------------


class NosPlugin(BaseModel):
    plugin: StrictStr
    configuration: Optional[Dict]


class ParamikoSshServerConfig(BaseModel):
    ssh_key_file: Optional[StrictStr] = None
    ssh_key_file_password: Optional[StrictStr] = None
    ssh_banner: Optional[StrictStr] = "FakeNOS Paramiko SSH Server"
    timeout: Optional[StrictInt] = 1
    address: Optional[Union[Literal["localhost"], IPvAnyAddress]]
    watchdog_interval: Optional[StrictInt] = 1


class ParamikoSshServerPlugin(BaseModel):
    plugin: Literal["ParamikoSshServer"]
    configuration: Optional[ParamikoSshServerConfig]


class CMDShellConfig(BaseModel):
    intro: Optional[StrictStr] = "Custom SSH Shell"
    ruler: Optional[StrictStr] = ""
    completekey: Optional[StrictStr] = "tab"
    newline: Optional[StrictStr] = "\r\n"


class CMDShellPlugin(BaseModel):
    plugin: Literal["CMDShell"]
    configuration: Optional[CMDShellConfig]


class InventoryDefaultSection(BaseModel):
    username: Optional[StrictStr]
    password: Optional[StrictStr]
    # port: Optional[Union[conint(strict=True, gt=0, le=65535), conlist(conint(strict=True, gt=0, le=65535), min_items=2, max_items=2, unique_items=True)]]
    # use this for now, mkdocstring having issue with pydantic - https://github.com/mkdocstrings/griffe/issues/66
    port: Optional[Union[StrictInt, List[StrictInt]]]
    server: Optional[Union[ParamikoSshServerPlugin]]
    shell: Optional[Union[CMDShellPlugin]]
    nos: Optional[NosPlugin]


class HostConfig(InventoryDefaultSection):
    # count: Optional[conint(strict=True, gt=0)]
    # use this for now, mkdocstring having issue with pydantic - https://github.com/mkdocstrings/griffe/issues/66
    count: Optional[StrictInt]

    @root_validator(pre=True)
    def check_port_value(cls, values):
        port = values.get("port")
        if "count" not in values and port:
            assert isinstance(
                port, int
            ), "If no host 'count' given, port must be an integer"
        elif "count" in values and port:
            assert isinstance(port, list), "If host 'count' given, port must be a list"
        return values


class model_fakenos_inventory(BaseModel):
    """FakeNOS inventory data schema"""

    default: Optional[InventoryDefaultSection]
    hosts: Dict[StrictStr, HostConfig]

    class Config:
        extra = "forbid"


# ---------------------------------------------------------------------------------------
# NOS plugin commands model
# ---------------------------------------------------------------------------------------

class model_nos_command(BaseModel):
    output: Union[StrictStr, None, Callable]
    help: Optional[StrictStr]
    prompt: Optional[Union[StrictStr, List[StrictStr]]]
    new_prompt: Optional[StrictStr]
    
class model_nos_attributes(BaseModel):
    commands: Dict[StrictStr, model_nos_command]
    name: StrictStr
    initial_prompt: StrictStr



import logging
import paramiko
import io
import threading
import time
import traceback

from fakenos.core.servers import TCPServerBase

log = logging.getLogger(__name__)

DEFAULT_SSH_KEY = """-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAnahBtR7uxtHmk5UwlFfpC/zxdxjUKPD8UpNOOtIJwpei7gaZ
+Jgub5GFJtTG6CK+DIZiR4tE9JxMjTEFDCGA3U4C36shHB15Pl3bLx+UxdyFylpc
c7XYp4fpQjhFUoHOAIl5ZaA223kIxi7sFXtM1Gjy6g49u+G5teVfMbeZnks2xjjy
F84qVADFBXCsfjrY5m4R+Wnfups/jP1agOpnOvqHlX/bpvzEZRcwJ0A8CylBZzQP
D1Y4EXy1B4QLyLJKFIMRkWnr0f8rK5Q/obCLTjl+IMmZrkItbfC/hYCy6TDi+Efn
cgGw02L93Mf6QGDNc21BsRELPYMME22MmpLphQIBIwKCAQEAmScbQjtOWr1GY3r7
/dG90SGaG+w70AALDmM2DUEQy6k/MF4vLAGMMd3RzfNE4YDV4EgHszbVRWSiIsHn
pWJf7OyyVZ7s9r2LuO111gFr82iB98V+YcaX8zOSIxIXdLicOwk0GZRSjA8tGErW
tcg8AYqFkulDSMylxqRN2IZ3+NnTROxh4uUFH57roSYoCvzjM2v1Xa+S42BLpBD1
3mLAJD36JhOhMTgYUgHAROx9+YUUUzYk3jpkTGWnAYSumnJXQYphLE9zadXxh94N
HZJdvXajuP5N2M3Q2b4Gbyt2wNFlNcHGA+Zwk8wHIBnY9Sb9Gz0QALsOAwUoRY8T
rCysSwKBgQDPVjFdSgM3jScmFV9fVnx3iNIlM6Ea7+UCrOOCvcGtzDo5vuTPktw7
8abHEFHw7VrtxI3lRQ41rlmK3B//Q7b+ZJ0HdZaRdyCqW1u91tq1tQe7yiJBm0c5
hZ3F0Vr6HAXoBVOux5wUq55jvUJ8dCVYNYfctZducVmOos3toDkSzQKBgQDCqRQ/
GO5AU3nKfuJ+SZvv8/gV1ki8pGmyxkSebUqZSXFx+rQEQ1e6tZvIz/rYftRkXAyL
XfzXX8mU1wEci6O1oSLiUBgnT82PtUxlO3Peg1W/cpKAaIFvvOIvUMRGFbzWhuj7
4p4KJjZWjYkAV2YlZZ8Br23DFFjjCuawX7NhmQKBgHCN4EiV5H09/08wLHWVWYK3
/Qzhg1fEDpsNZZAd3isluTVKXvRXCddl7NJ2kuHf74hjYvjNt0G2ax9+z4qSeUhF
P00xNHraRO7D4VhtUiggcemZnZFUSzx7vAxNFCFfq29TWVBAeU0MtRGSoG9yQCiS
Fo3BqfogRo9Cb8ojxzYXAoGBAIV7QRVS7IPheBXTWXsrKRmRWaiS8AxTe63JyKcm
XwoGea0+MkwQ67M6s/dqCxgcdGITO81Hw1HbSGYPxj91shYlWb/B5K0+CUyZk3id
y8vHxcUbXSTZ8ls/sQqAhpZ1Tkn2HBpvglAaM+OUQK/G5vUSe6liWeTawJuvtCEr
rjRLAoGAUNNY4/7vyYFX6HkX4O2yL/LZiEeR6reI9lrK/rSA0OCg9wvbIpq+0xPG
jCrc8nTlA0K0LtEnE+4g0an76nSWUNiP4kALROfZpXajRRaWdwFRAO17c9T7Uxc0
Eez9wYRqHiuvU0rryYvGyokr62w1MtJO0tttnxe1Of6wzb1WeCU=
-----END RSA PRIVATE KEY-----"""


class ParamikoSshServerInterface(paramiko.ServerInterface):
    def __init__(
        self,
        ssh_banner="FakeNOS Paramiko SSH Server",
        username=None,
        password=None,
    ):
        self.ssh_banner = ssh_banner
        self.username = username
        self.password = password

    def check_channel_request(self, kind, chanid):
        """
        This will allow the SSH server to provide a channel for the client
        to communicate over. By default, this will return
        OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED, so  we have to override it
        to return OPEN_SUCCEEDED when the kind of channel requested is "session".
        """
        if kind == "session":
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_channel_pty_request(
        self, channel, term, width, height, pixelwidth, pixelheight, modes
    ):
        """
        AFAIK, pty (pseudo-tty (TeleTYpewriter)) will allow our client to interact
        with our shell.
        """
        return True

    def check_channel_shell_request(self, channel):
        """This allows us to provide the channel with a shell we can connect to it."""
        return True

    def check_auth_password(self, username, password):
        if (username == self.username) and (password == self.password):
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def get_banner(self):
        """
        String that will display when a client connects, before authentication has
        happened. This is different than the shell's intro property, which is displayed
        after the authentication.
        """
        return (self.ssh_banner + "\r\n", "en-US")


class TapIO(io.StringIO):
    """
    Class to implement StringIO subclass but with blocking readline method
    and a list to buffer lines on write
    """

    def __init__(self, run_srv, initial_value="", newline="\n"):
        self.lines = []
        self.run_srv = run_srv
        super(TapIO, self).__init__(initial_value, newline)

    def readline(self, size=-1):
        """method to readline in indefinite block mode"""
        while self.run_srv.is_set():
            if self.lines:
                return self.lines.pop(-1)
            time.sleep(0.01)

    def write(self, value):
        """
        :param value: line to add to self.lines buffer
        """
        self.lines.insert(0, value)


def channel_to_shell_tap(channel_stdio, shell_stdin, shell_replied_event, run_srv):
    buffer = io.BytesIO()
    while run_srv.is_set():
        byte = channel_stdio.read(1)
        if byte in (b"\r", b"\n"):
            # in case several \n received from channel need to
            # make sure to wait for shell to process replies
            shell_replied_event.wait(10)
            # echo input back to the client
            channel_stdio.write("\r\n")
            # read line from buffer, clear buffer, send line to cmd shell
            buffer.write(b"\r\n")
            buffer.seek(0)
            line = buffer.read().decode(encoding="utf-8")
            buffer.seek(0)
            buffer.truncate()
            log.debug(
                "ssh_server.channel_to_shell_tap sending line to shell: {}".format(
                    [line]
                )
            )
            shell_stdin.write(line)
            shell_replied_event.clear()
        else:
            shell_replied_event.wait(10)
            # echo input back to the client
            try:
                channel_stdio.write(byte)
            except OSError as e:
                log.error(
                    "ssh_server.channel_to_shell_tap channel write error: {}".format(e)
                )
                # do nothing, exit, watcdog will handle termination
                break
            # safe received character to buffer
            buffer.write(byte)
        time.sleep(0.01)


def shell_to_channel_tap(channel_stdio, shell_stdout, shell_replied_event, run_srv):
    while run_srv.is_set():
        line = shell_stdout.readline()
        # line is None we break the loop
        if line is None:
            break
        if "\r\n" not in line and "\n" in line:
            line = line.replace("\n", "\r\n")
        log.debug(
            "ssh_server.shell_to_channel_tap sending line to channel {}".format([line])
        )
        channel_stdio.write(line.encode(encoding="utf-8"))
        shell_replied_event.set()


class ParamikoSshServer(TCPServerBase):
    def __init__(
        self,
        shell,
        nos,
        port,
        username,
        password,
        ssh_key_file=None,
        ssh_key_file_password=None,
        ssh_banner="FakeNOS Paramiko SSH Server",
        shell_configuration=None,
        address="127.0.0.1",
        timeout=1,
        watchdog_interval=1,
    ):
        super(ParamikoSshServer, self).__init__()

        self.nos = nos
        self.shell = shell
        self.shell_configuration = shell_configuration or {}
        self.ssh_banner = ssh_banner
        self.username = username
        self.password = password
        self.port = port
        self.address = address
        self.timeout = timeout
        self.watchdog_interval = watchdog_interval

        if ssh_key_file:
            self._ssh_server_key = paramiko.RSAKey.from_private_key_file(
                ssh_key_file, ssh_key_file_password
            )
        else:
            self._ssh_server_key = paramiko.RSAKey(
                file_obj=io.StringIO(DEFAULT_SSH_KEY)
            )

    def watchdog(self, run_srv, session, shell):
        """
        Method to monitor server liveness and recover where possible.
        """
        while run_srv.is_set():
            # check if session is alive, stop shell if it is not
            if not session.is_alive():
                log.warning(
                    "ParamikoSshServer.watchdog - session not alive, stopping shell"
                )
                shell.stop()
                break

            time.sleep(self.watchdog_interval)

    def connection_function(self, client):
        shell_replied_event = threading.Event()
        run_srv = threading.Event()
        run_srv.set()

        # create the SSH transport object
        session = paramiko.Transport(client)
        session.add_server_key(self._ssh_server_key)

        # create the server
        server = ParamikoSshServerInterface(
            ssh_banner=self.ssh_banner,
            username=self.username,
            password=self.password,
        )

        # start the SSH server
        session.start_server(server=server)

        # create the channel and get the stdio
        channel = session.accept()
        channel_stdio = channel.makefile("rw")

        # create stdio for the shell
        shell_stdin, shell_stdout = TapIO(run_srv), TapIO(run_srv)

        # start intermediate thread to tap into the channel_stdio->shell_stdin bytes stream
        channel_to_shell_tapper = threading.Thread(
            target=channel_to_shell_tap,
            args=(channel_stdio, shell_stdin, shell_replied_event, run_srv),
        )
        channel_to_shell_tapper.start()

        # start intermediate thread to tap into the shell_stdout->channel_stdio bytes stream
        shell_to_channel_tapper = threading.Thread(
            target=shell_to_channel_tap,
            args=(channel_stdio, shell_stdout, shell_replied_event, run_srv),
        )
        shell_to_channel_tapper.start()

        # create the client shell
        client_shell = self.shell(
            stdin=shell_stdin,
            stdout=shell_stdout,
            nos=self.nos,
            is_running=self._is_running,
            **self.shell_configuration,
        )

        # start watchdog thread
        watchdog_thread = threading.Thread(
            target=self.watchdog, args=(run_srv, session, client_shell)
        )
        watchdog_thread.start()

        # running this command will block this function until shell exits
        client_shell.start()
        log.debug("ParamikoSshServer.connection_function stopped shell thread")

        # kill this server threads - watchdog, TapIO, shell_to_channel_tapper and channel_to_shell_tapper
        run_srv.clear()
        log.debug("ParamikoSshServer.connection_function stopped server threads")

        # After execution continues, we can close the session
        session.close()
        log.debug(f"ParamikoSshServer.connection_function closed transport {session}")



from fakenos.plugins.servers import servers_plugins
from fakenos.plugins.nos import nos_plugins
from fakenos.plugins.shell import shell_plugins
from fakenos.core.host import Host
from fakenos.core.nos import Nos
from fakenos.core.pydantic_models import model_fakenos_inventory

from typing import (
    Union,
    Callable,
    Dict,
    List,
)
import logging
import fnmatch
import copy
import yaml

log = logging.getLogger(__name__)

default_inventory = {
    "default": {
        "username": "user",
        "password": "user",
        "port": [10000, 60000],
        "server": {
            "plugin": "ParamikoSshServer",
            "configuration": {
                "address": "0.0.0.0",
                "timeout": 1,
            },
        },
        "shell": {"plugin": "CMDShell", "configuration": {}},
        "nos": {"plugin": "cisco_ios", "configuration": {}},
    },
    "hosts": {
        "router1": {"port": 6001},
        "router2": {"port": 6002},
    },
}


class FakeNOS:
    """
    FakeNOS class is a main entry point to interact with fake NOS servers - start, stop, list.

    :param inventory: FakeNOS inventory dictionary or OS path to .yaml file with inventory data
    :param log_level: logging level to use

    Sample usage:

    ```python
    from fakenos import FakeNOS

    net = FakeNOS()
    net.start()
    ```
    """

    def __init__(self, inventory: dict = None, log_level: str = "DEBUG") -> None:
        self.inventory = inventory or default_inventory
        self.hosts = {}
        self.allocated_ports = set()
        self.shell_plugins = shell_plugins
        self.nos_plugins = nos_plugins
        self.servers_plugins = servers_plugins
        self.log_level = log_level

        self._configure_logging()
        self._load_inventory()
        self.init()

    def _configure_logging(self) -> None:
        """Helper method to setup logging"""
        logging.basicConfig(level=self.log_level.upper())

    def _load_inventory(self) -> None:
        """Helper method to load FakeNOS inventory"""
        # load yaml inventory
        if isinstance(self.inventory, str) and self.inventory.endswith(".yaml"):
            with open(self.inventory, "r", encoding="utf-8") as f:
                self.inventory = yaml.safe_load(f.read())

        # make sure we have defaults
        self.inventory["default"] = {
            **default_inventory["default"],
            **self.inventory.get("default", {}),
        }

        # validate inventory data
        inventory_model_instance = model_fakenos_inventory(**self.inventory)
        log.debug("FakeNOS inventory validation succeeded")
        # log.debug(str(inventory_model_instance.schema_json(indent=4)))

    def init(self) -> None:
        """
        Helper method to initiate host objects and store them in self.hosts, this
        method called automatically on FakeNOS object instantiation.
        """
        for host, host_config in self.inventory["hosts"].items():
            params = {
                **copy.deepcopy(self.inventory["default"]),
                **copy.deepcopy(host_config),
            }
            port = params.pop("port")
            count = params.pop("count", None)
            if count:
                for i in range(0, count):
                    name = f"{host}{i+1}"
                    port_ = self._allocate_port(port)
                    self.hosts[name] = Host(
                        name=name, port=port_, fakenos=self, **copy.deepcopy(params)
                    )
            else:
                port_ = self._allocate_port(port)
                self.hosts[host] = Host(name=host, port=port_, fakenos=self, **params)

    def _allocate_port(self, port: int) -> None:
        """
        Method to allocate port for host

        :param port: integer or list of two integers - range to allocate port from
        """
        if isinstance(port, int):
            if port in self.allocated_ports:
                raise ValueError(f"Port {port} already in use")
            allocated_port = port
        elif isinstance(port, list):
            for p in range(port[0], port[1] + 1):
                if p not in self.allocated_ports:
                    allocated_port = p
                    break
            else:
                raise RuntimeError("Port allocation failed")
        else:
            raise TypeError(
                "Unsupported port type {}, supported int or list".format(type(port))
            )

        self.allocated_ports.add(allocated_port)

        return allocated_port

    def _split_pattern(self, pattern: Union[str, List[str]]) -> List[str]:
        """
        Helper method to split pattern into a list of patterns.

        :param pattern: glob pattern or list or comma separated list of patterns
        :return: list of patterns
        """
        return (
            pattern
            if isinstance(pattern, list)
            else [i.strip() for i in pattern.split(",")]
        )

    def start(self, hosts: Union[str, List[str]] = "*") -> None:
        """
        Function to start NOS servers instances

        :param hosts: glob pattern to match hosts to start by their name or
            list or comma separated list of patterns
        """
        hosts = self._split_pattern(hosts)
        for h in self.hosts.values():
            if not h.running and any(fnmatch.fnmatchcase(h.name, p) for p in hosts):
                h.start()

    def stop(self, hosts: Union[str, List[str]] = "*") -> None:
        """
        Function to stop NOS servers instances

        :param hosts: glob pattern to match hosts to stop by their name or
            list or comma separated list of patterns
        """
        hosts = self._split_pattern(hosts)
        for h in self.hosts.values():
            if h.running and any(fnmatch.fnmatchcase(h.name, p) for p in hosts):
                h.stop()

    def register_nos_plugin(self, plugin: Union[str, Dict, Nos]) -> None:
        """
        Method to register NOS plugin with FakeNOS object, all plugins
        must be registered before calling start method.

        :param plugin: OS path string to NOS plugin `.yaml/.yml` or `.py` file,
          dictionary or instance if Nos class
        """
        if isinstance(plugin, Nos):
            nos_instance = plugin
        else:
            nos_instance = Nos()
            if isinstance(plugin, dict):
                nos_instance.from_dict(plugin)
            elif isinstance(plugin, str):
                nos_instance.from_file(plugin)
            else:
                raise TypeError(
                    "Unsupported NOS type {}, supported str, dict or Nos".format(
                        type(plugin)
                    )
                )
        self.nos_plugins[nos_instance.name] = nos_instance

    def list_hosts(self, hosts: Union[str, List[str]] = "*") -> list:
        """
        Method to produce a list of hosts wit inventory and status information

        :param hosts: glob pattern to match hosts to return by their name or
            list or comma separated list of patterns
        """
        ret = []
        hosts = self._split_pattern(hosts)
        for h in self.hosts.values():
            if any(fnmatch.fnmatchcase(h.name, p) for p in hosts):
                ret.append(
                    {
                        "name": h.name,
                        "nos": h.nos_params["plugin"],
                        "shell": h.shell_params["plugin"],
                        "server": h.server_params["plugin"],
                        "port": h.server.port if h.server else h.port,
                        "address": h.server.address if h.server else None,
                        "base_prompt": h.shell_params["configuration"]["base_prompt"],
                        "running": h.running,
                    }
                )
        return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fakenos',
 'fakenos.core',
 'fakenos.plugins',
 'fakenos.plugins.nos',
 'fakenos.plugins.servers',
 'fakenos.plugins.shell',
 'fakenos.plugins.tapes',
 'fakenos.plugins.utils']

package_data = \
{'': ['*']}

install_requires = \
['paramiko>=2.10.3,<3.0.0', 'pydantic>=1.9.0,<2.0.0', 'pyyaml>=6.0.0,<7.0.0']

entry_points = \
{'console_scripts': ['fakenos = fakenos.plugins.utils.cli:run_cli']}

setup_kwargs = {
    'name': 'fakenos',
    'version': '0.1.0',
    'description': 'Fake Network Operating System',
    'long_description': '# Fake Network Operating Systems - FakeNOS\n\n> "Reality is merely an illusion, albeit a very persistent one."\n>\n> ~ Albert Einstein\n\nFakeNOS created to simulate Network Operating Systems interactions.\n\n[Documentation](https://dmulyalin.github.io/fakenos/)\n\n## FakeNOS inspired by and borrowed from\n\n- [sshim](https://pythonhosted.org/sshim/) - library for testing and debugging SSH automation clients\n- [PythonSSHServerTutorial](https://github.com/ramonmeza/PythonSSHServerTutorial) - tutorial on creating paramiko based SSH server\n- [fake-switches](https://github.com/internap/fake-switches) - pluggable switch/router command-line simulator\n- [ncs-netsim](https://developer.cisco.com/docs/nso/guides/#!the-network-simulator) - tool to simulate a network of devices\n- [cisshgo](https://github.com/tbotnz/cisshgo) - concurrent SSH server to emulate network equipment for testing purposes\n- [scrapli-replay](https://pypi.org/project/scrapli-replay/) - tools to enable easy testing of SSH programs and to create semi-interactive SSH servers\n',
    'author': 'Denis Mulyalin',
    'author_email': 'd.mulyalin@gmail.com',
    'maintainer': 'Denis Mulyalin',
    'maintainer_email': 'd.mulyalin@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.5,<3.10',
}


setup(**setup_kwargs)



from cmd import Cmd
import logging
import time
import traceback

log = logging.getLogger(__name__)


class CMDShell(Cmd):
    """"""

    use_rawinput = False

    commands = {
        "exit": {"output": True, "help": "Exit commands shell"},
        "_default_": {
            "output": "Unknown command",
            "help": "Output to print for unknown commands",
        },
    }

    def __init__(
        self,
        stdin,
        stdout,
        nos,
        base_prompt,
        is_running,
        intro="Custom SSH Shell",
        ruler="",
        completekey="tab",
        newline="\r\n",
    ):
        self.nos = nos
        self.ruler = ruler
        self.intro = intro
        self.base_prompt = base_prompt
        self.commands.update(nos.commands or {})
        self.newline = newline
        self.prompt = nos.initial_prompt.format(base_prompt=base_prompt)
        self.is_running = is_running

        # call the base constructor of cmd.Cmd, with our own stdin and stdout
        super(CMDShell, self).__init__(
            completekey=completekey,
            stdin=stdin,
            stdout=stdout,
        )

    def start(self):
        self.cmdloop()

    def stop(self):
        self.stdin.write("exit" + self.newline)

    def writeline(self, value):
        for line in str(value).splitlines():
            self.stdout.write(line + self.newline)

    def emptyline(self):
        """This method to do nothing if empty line entered"""
        pass

    def precmd(self, line):
        return line

    def postcmd(self, stop, line):
        return stop

    def do_help(self, *args):
        """Method to return help for commands"""
        lines = {}  # dict of {cmd: cmd_help}
        width = 0  # record longest command width for padding
        cmd_string = " ".join(args)
        # form help for all commands
        for cmd, cmd_data in self.commands.items():
            # skip special commands
            if cmd.startswith("_") and cmd.endswith("_"):
                continue
            # skip commands that does not match current prompt
            if not self._check_prompt(cmd_data.get("prompt")):
                continue
            lines[cmd] = cmd_data.get("help", "")
            width = max(width, len(cmd))
        # form help lines
        help_msg = []
        for k, v in lines.items():
            padding = " " * (width - len(k)) + "  "
            help_msg.append(f"{k}{padding}{v}")
        self.writeline(self.newline.join(help_msg))

    def _check_prompt(self, prompt_):
        """
        Helper method to check if prompt_ matches current prompt

        :param prompt_: (string or None)  prompt to check
        """
        # prompt_ is None if no 'prompt' key defined for command
        if prompt_ is None:
            return True
        if isinstance(prompt_, str):
            return self.prompt == prompt_.format(base_prompt=self.base_prompt)
        if isinstance(prompt_, list):
            return any(
                self.prompt == i.format(base_prompt=self.base_prompt) for i in prompt_
            )

    def default(self, command):
        """Method called if no do_xyz methods found"""
        log.debug("shell.run_command running command '{}'".format(command))
        ret = self.commands["_default_"]["output"]
        try:
            cmd_data = self.commands[command]
            if self._check_prompt(cmd_data.get("prompt")):
                ret = cmd_data["output"]
                if callable(ret):
                    ret = ret(
                        base_prompt=self.base_prompt,
                        current_prompt=self.prompt,
                        command=command,
                    )
                if "new_prompt" in cmd_data:
                    self.prompt = cmd_data["new_prompt"].format(
                        base_prompt=self.base_prompt
                    )
        except KeyError:
            pass
        except:
            ret = traceback.format_exc()
            ret = ret.replace("\n", self.newline)

        # returning True will close the shell
        if ret is True:
            return True
        # if not None, write output to stdout
        elif ret is not None:
            self.writeline(ret)



"""
Host classes
"""
from typing import Union, Callable, Dict, List
import logging

log = logging.getLogger(__name__)


class Host:
    def __init__(
        self,
        name: str,
        username: str,
        password: str,
        port: int,
        server: dict,
        shell: dict,
        nos: dict,
        fakenos,
    ) -> None:
        self.name = name
        self.server_params = server
        self.shell_params = shell
        self.nos_params = nos
        self.username = username
        self.password = password
        self.port = port
        self.fakenos = fakenos
        self.shell_params["configuration"].setdefault("base_prompt", self.name)
        self.running = False
        self.server = None
        self.server_plugin = None
        self.shell_plugin = None
        self.nos_plugin = None

    def start(self):
        """Method to start server instance for this hosts"""
        self.server_plugin = self.fakenos.servers_plugins[self.server_params["plugin"]]
        self.shell_plugin = self.fakenos.shell_plugins[self.shell_params["plugin"]]
        self.nos_plugin = self.fakenos.nos_plugins[self.nos_params["plugin"]]

        self.server = self.server_plugin(
            shell=self.shell_plugin,
            shell_configuration=self.shell_params["configuration"],
            nos=self.nos_plugin,
            port=self.port,
            username=self.username,
            password=self.password,
            **self.server_params["configuration"],
        )
        self.server.start()
        self.running = True

    def stop(self):
        """Method to stop server instance of this host"""
        self.server.stop()
        self.server = None
        self.running = False
